#!/bin/bash
set -eux

# This script will fetch binaries and create resource tarballs for use by
# charm-[push|release]. The arm64 binaries are not available upsteram for
# v2.6, so we must build them and host them somewhere ourselves. The steps
# for doing that are documented here:
#
# https://gist.github.com/kwmonroe/9b5f8dac2c17f93629a1a3868b22d671

# Supported calico architectures
arches="amd64 arm64"
calico_cni_version="v3.4.0"

function fetch_and_validate() {
  # fetch a binary and make sure it's what we expect (executable > 20MB)
  min_bytes=20000000
  location="${1-}"
  if [ -z ${location} ]; then
    echo "$0: Missing location parameter for fetch_and_validate"
    exit 1
  fi

  # remove everything up until the last slash to get the filename
  filename=$(echo "${location##*/}")
  case ${location} in
    http*)
      fetch_cmd="wget ${location} -O ./${filename}"
      ;;
    *)
      fetch_cmd="scp ${location} ./${filename}"
      ;;
  esac
  ${fetch_cmd}

  # Make sure we fetched something big enough
  actual_bytes=$(wc -c < ${filename})
  if [ $actual_bytes -le $min_bytes ]; then
    echo "$0: ${filename} should be at least ${min_bytes} bytes"
    exit 1
  fi

  # Make sure we fetched a binary
  if ! file ${filename} 2>&1 | grep -q 'executable'; then
    echo "$0: ${filename} is not an executable"
    exit 1
  fi
}

for arch in ${arches}; do
  rm -rf resource-build-$arch
  mkdir resource-build-$arch
  pushd resource-build-$arch
  fetch_and_validate \
    https://github.com/projectcalico/cni-plugin/releases/download/$calico_cni_version/calico-$arch
  fetch_and_validate \
    https://github.com/projectcalico/cni-plugin/releases/download/$calico_cni_version/calico-ipam-$arch
  mv calico-$arch calico
  mv calico-ipam-$arch calico-ipam
  echo "calico cni-plugin $calico_cni_version" > BUILD_INFO

  chmod +x calico calico-ipam
  tar -zcvf ../calico-cni-$arch.tar.gz .

  popd
  rm -rf resource-build-$arch
done
